/*
 * FB Alpha Neo Geo module
 *
 * There are 264 scanlines. The image is 224 lines high, starts at line 24, and ends at line 248.
 * The IRQ position register can behave differently when it's written at line 248 or higher;
 *  - If bit 5 is set:
 *     - If the line is < 16 (???), the offset is the top of the screen (mslug2) or vblank (lblade2);
 *     - If the line is > 15 (???), the normal scanline offset is used.
 *
 * The memory card interface occupies 32KB of memory. Only odd bytes are used to address memory.
 * The BIOS tests the size of the inserted card by writing to the first byte of each 2KB block.
 * There are some registers at the start of the address space (banking and other features).
 * The inteface used is JEIDA 3.0
 *
 * bit 6 of memory location 0x300081 = 0						 -> 1/2 slot MVS
 *									 = 1 + bit 5 of 0x320001 = 0 -> 4 	slot MVS
 *									 						 = 1 -> 6   slot MVS
 */

#include "neogeo.h"
#include "burn_ym2610.h"
#include "bitswap.h"

#define Z80_SWITCH_ROMBANK_ON_RESET
#define IRQ_KLUDGE

#ifdef IRQ_KLUDGE
 static int nIRQOffset;
#else
 static const int nIRQOffset = 14;
#endif

#ifdef IRQ_KLUDGE
 static unsigned short nScanlineOffset;
#else
 // 0xF8 is correct as verified on MVS hardware
 static const unsigned short nScanlineOffset = 0xF8;
#endif

static int nCyclesScanline = 12000000 / (60 * 264);

// Which 68K BIOS to use
static int nBIOS;

unsigned char NeoButton1[32] = { 0, };
unsigned char NeoButton2[8]  = { 0, };
unsigned char NeoButton3[8]  = { 0, };
unsigned char NeoButton4[8]  = { 0, };
unsigned char NeoJoy1[8]	 = { 0, };
unsigned char NeoJoy2[8]     = { 0, };
unsigned char NeoJoy3[8]     = { 0, };
unsigned char NeoJoy4[8]     = { 0, };
unsigned short NeoAxis[2]	 = { 0, };
unsigned char NeoInput[32]   = { 0, };
unsigned char NeoDiag[2]	 = { 0, };
unsigned char NeoReset = 0, NeoSystem = 0;

// Joyports are multiplexed
static int nJoyport0[8] = { 0, };
static int nJoyport1[8] = { 0, };

// Ports always mirror the joystick, except when other controllers are hooked up:
//
// Joyport0: 0x00		 : trackball X in irrmaze / p1 paddle in popbounc
//			 0x01		 : trackball Y in irrmaze
//			 0x09		 : mahjong controller
//			 0x12		 : mahjong controller (BIOS controls)
//			 0x1B		 : Always has p1 joystick & buttons
//			 0x24		 : mahjong controller
//			 0x20 & 0x21 : read by irrmaze instead of 0x00 & 0x01 when you lose a life
//						   (activates air-jets)
// Joyport1: 0x00		 : p2 paddle in popbounc
//			 0x1B		 : Always has p2 joystick & buttons

// ----------------------------------------------------------------------------
// Variables that need to be included in savestates

bool bNeoEnableGraphics;

unsigned int nNeo68KROMBank;

static int nIRQAcknowledge;

static int nIRQ2Control;
static bool bSRAMWritable;

static bool b68KBoardROMBankedIn;
static bool bZ80BoardROMBankedIn;
static int nZ80Bank0, nZ80Bank1, nZ80Bank2, nZ80Bank3;

static unsigned char* NeoGraphicsRAMBank;
static unsigned short NeoGraphicsRAMPointer;
static int nNeoGraphicsModulo;

static int nSpriteFrameSpeed;
static int nSpriteFrameTimer;

static unsigned char nSoundLatch;
static unsigned char nSoundReply;
static unsigned int nSoundStatus;
static int nSoundPrevReply;

static int nInputSelect;
static unsigned char* NeoInputBank;
static unsigned int nAnalogAxis[2];

static unsigned int nuPD4990ATicks;

static int nIRQLine, nIRQLineLo, nIRQLineHi;

static int nSMARandomNumber;

// ----------------------------------------------------------------------------

static int nSRAMProtAddress;
int nNeoSRAMProtection = -1;
int ncthd2003 = -1;
int nNeoPCM2PLAYMOREByteswap = 0; // PCM2 PLAYMORE chip
int nNeoPCM2SNKByteswap = 0; // Prevents errors when loading a non-NEO-PCM2 game
int nADPCMByteswap = 0; // Simple byteswap for svcboot and clones/matrim and clones
int nNeoTextROMFixType = 0; // Text fix for games that have TextBankswitch
// Decrypted ROMs saving functions Thanks to IQ_132
int bsavedecrypteds1 = 0;
int bsavedecryptedvs = 0;
int bsavedecryptedps = 0;
int bsavedecryptedcs = 0;

static bool bMemoryCardInserted, bMemoryCardWritable;

unsigned int nNeoSMARNGAddress[2] = { 0, 0 };

void (*pNeoInitCallback)() = NULL;

pSekWriteWordHandler pNeoSMABankswitch;

static int nCyclesDone[2], nCyclesTotal[2];
static int nCyclesSegment;
static int nCurrentCPU;
static int nNeoCPUSpeedfactor;

static unsigned char *AllRAM = NULL, *RAMEnd = NULL, *AllROM = NULL, *ROMEnd = NULL;

unsigned char *Neo68KROM01, *Neo68KROM02;
unsigned char *NeoZ80ROM;
static unsigned char *Neo68KBIOS, *Neo68KVectors, *NeoZ80BIOS;
static unsigned char *Neo68KRAM, *NeoZ80RAM, *NeoNVRAM, *NeoNVRAM2, *NeoMemoryCard;

unsigned int nSpriteSize;
static unsigned int nCodeSize;

unsigned char* NeoGraphicsRAM;

unsigned char* YM2610ADPCMAROM = NULL;
unsigned char* YM2610ADPCMBROM = NULL;

static int nYM2610ADPCMASize = 0;
static int nYM2610ADPCMBSize = 0;

static int nTriggerIRQ;
static bool bIRQ2Enabled;

// static bool bAllowRasters = false;

static bool bRenderImage;

static bool bRenderLineByLine;

static bool bForcePartialRender;
static bool bForceUpdateOnStatusRead;

static int nNeoControlConfig;

static bool bAESBIOS;

// This function is called once to determine how much memory is needed (RAMEnd-(unsigned char *)0),
// then a second time after the memory is allocated to set up all the pointers.
static int RAMIndex()
{
	unsigned char* Next; Next = AllRAM;
	NeoPalSrc[0]	= Next; Next += 0x002000;			// Palette RAM Bank 0
	NeoPalSrc[1]	= Next; Next += 0x002000;			// Palette RAM Bank 1

	NeoGraphicsRAM	= Next; Next += 0x020000;			// Graphics controller RAM (2 64KB banks)

	Neo68KRAM		= Next; Next += 0x010000;			// 68K work RAM
	NeoZ80RAM		= Next; Next += 0x000800;			// Z80 RAM

	NeoNVRAM		= Next; Next += 0x010000;			// Battery backed SRAM
	if ((BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK) == HARDWARE_SNK_GAMBLING) {
		NeoNVRAM2	= Next; Next += 0x010000;			// Extra SRAM for vliner/jockeygp
	}
	NeoMemoryCard	= Next; Next += 0x004000;			// Memory card

	RAMEnd			= Next;

	return 0;
}

// This function is called once to determine how much memory is needed (ROMEnd-(unsigned char*)0),
// then a second time after the memory is allocated to set up all the pointers.
static int ROMIndex()
{
	unsigned char* Next; Next = AllROM;
	Neo68KBIOS		= Next; Next += 0x040000;			// 68K boardROM
	Neo68KVectors	= Next; Next += 0x000400;			// Copy of 68K cartridge ROM with boardROM vector table

	Neo68KROM01 = Next; Next += nCodeSize;
	if (nCodeSize > 0x100000) {
		// Bankable part of 68K cartridge ROM
		Neo68KROM02	= Neo68KROM01 + 0x100000;
	} else {
		// Mirror of the 68K cartridge ROM for smaller games
		Neo68KROM02	= Neo68KROM01;
	}

	NeoZ80BIOS		= Next; Next += 0x020000;			// Z80 boardROM
	NeoZ80ROM		= Next; Next += 0x080000;

	NeoZoomROM		= Next; Next += 0x010000;			// Y Zoom table

	ROMEnd			= Next;

	return 0;
}

// -----------------------------------------------------------------------------
// ROM loading

static inline bool NeoCheckAESBIOS()
{
	if (nBIOS == -1 || nBIOS == 6) {
		return true;
	}

	return false;
}

static int NeoLoad68KBIOS(int nNewBIOS)
{
	switch (NeoSystem & 0x30) {
		case 0x00: {
			switch (nNewBIOS) {
				// MVS
				case 0x00:
				case 0x01:
				case 0x02:
				// AES
				case 0x06:
					break;
				default:
					nNewBIOS = -1;
			}
			break;
		}
		case 0x10: {
			nNewBIOS = 0x08;
			break;
		}
		case 0x20: {
			nNewBIOS = 0x0C;
			break;
		}
		default: {
			nNewBIOS = -1;
		}
	}

	if ((BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK) == HARDWARE_SNK_TRACKBALL) {
		nNewBIOS = 9;
	}

	// Check if we need to load a new BIOS
	if (nNewBIOS == nBIOS) {
		return 0;
	}

	nBIOS = nNewBIOS;

	// Load the BIOS ROMs
	if (nBIOS >= 0) {
		BurnLoadRom(Neo68KBIOS, 0x00080 + nBIOS, 1);
	} else {
		BurnLoadRom(Neo68KBIOS, 0x00080 +     6, 1);
	}

	// Patch out checksum test (the AES BIOS doesn't have one)
	if (nBIOS == 9) {
		*((unsigned short*)(Neo68KBIOS + 0x010D8C)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x010D8E)) = 0x4E71;
	}

	// Hacks for Neo-Geo BIOS
	// Patch out the AS (europe), MVS, 1 slot, version 2 BIOS
	if (nBIOS == 2) {
		/* Remove memory check for now */
		*((unsigned short*)(Neo68KBIOS + 0x011B00)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x011B02)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x011B16)) = 0x4EF9;
		*((unsigned short*)(Neo68KBIOS + 0x011B18)) = 0x00C1;
		*((unsigned short*)(Neo68KBIOS + 0x011B1A)) = 0x1B6A;

		/* Patch bios rom, for Calendar errors */
		*((unsigned short*)(Neo68KBIOS + 0x011C14)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x011C16)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x011C1C)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x011C1E)) = 0x4E71;

		/* Patch checksum test */
		*((unsigned short*)(Neo68KBIOS + 0x011C62)) = 0x4E71;
		*((unsigned short*)(Neo68KBIOS + 0x011C64)) = 0x4E71;
	}

	// Create copy of 68K with BIOS vector table
	memcpy(Neo68KVectors + 0x00, Neo68KBIOS,		 0x0080);
	memcpy(Neo68KVectors + 0x80, Neo68KROM01 + 0x80, 0x0380);

	return 0;
}

static int FindType(const char* pName)
{
	int i = 0;

	while (pName[i] && pName[i] != '-' && pName[i] != '_') {
		i++;
	}

	return i + 1;
}

static int FindROMs(const char* nType, int* pOffset, int* pNum)
{
	int nOffset = -1;
	int nNum = -1;
	char* pName = NULL;

	// Invalidate indices
	if (pOffset) {
		*pOffset = -1;
	}
	if (pNum) {
		*pNum = 0;
	}

	if (nType[0] == 'p') {
		struct BurnRomInfo ri;
		ri.nType = 0;
		ri.nLen = 0;

		do {
			BurnDrvGetRomInfo(&ri, ++nOffset);
		} while (ri.nType != 0x10 && nOffset < 0x80);

		if (nOffset >= 0x7F) {
			return 1;
		}

		do {
			BurnDrvGetRomInfo(&ri, nOffset + ++nNum);
		} while (ri.nType == 0x10);

		if (pOffset) {
			*pOffset = nOffset;
		}
		if (pNum) {
			*pNum = nNum;
		}
	} else {
		int nTypeOffset;
		if (nType[1] == ' ') {

			// Skip nnn-sma.bin ROM
			if (nType[0] == 's') {
				BurnDrvGetRomName(&pName, ++nOffset, 0);
				nTypeOffset = FindType(pName);
				if (pName[nTypeOffset + 0] == 's'&& pName[nTypeOffset + 1] == 'm' && pName[nTypeOffset + 2] == 'a') {
					nOffset++;
				}
			}

			do {
				BurnDrvGetRomName(&pName, ++nOffset, 0);
				nTypeOffset = FindType(pName);
			} while (pName[nTypeOffset + 0] != nType[0] && nOffset < 0x80);
		} else {

			// 'V' ROMs can have 2 numbers; there is either: Vn = ADPCMA only, or: V1n = ADPCMA + V2n = ADPCMB

			do {
				BurnDrvGetRomName(&pName, ++nOffset, 0);
				nTypeOffset = FindType(pName);
			} while ((pName[nTypeOffset + 0] != nType[0] || pName[nTypeOffset + 1] != nType[1] || pName[nTypeOffset + 2] < '1' || pName[nTypeOffset + 2] > '9') && nOffset < 0x80);
		}

		if (nOffset >= 0x7F) {
			return 1;
		}

		if (nType[1] == ' ') {
			do {
				BurnDrvGetRomName(&pName, nOffset + ++nNum, 0);
				nTypeOffset = FindType(pName);
			} while (pName[nTypeOffset + 0] == nType[0] && nOffset < 0x80);
		} else {

			// 'V' ROMs can have 2 numbers; there is either: Vn = ADPCMA only, or: V1n = ADPCMA + V2n = ADPCMB

			do {
				BurnDrvGetRomName(&pName, nOffset + ++nNum, 0);
				nTypeOffset = FindType(pName);
			} while (pName[nTypeOffset + 0] == nType[0] && pName[nTypeOffset + 1] == nType[1] && pName[nTypeOffset + 2] >= '1' && pName[nTypeOffset + 2] <= '9' && nOffset < 0x80);
		}

		if (pOffset) {
			*pOffset = nOffset;
		}
		if (pNum) {
			*pNum = nNum >= 0 ? nNum : 0;
		}
	}

	return 0;
}

static int LoadRoms(NeoGameInfo* pInfo)
{
	NeoSpriteROM = (unsigned char*)malloc(nSpriteSize);
	if (NeoSpriteROM == NULL) {
		return 1;
	}

	if (BurnDrvGetHardwareCode() & (HARDWARE_SNK_ENCRYPTED_A | HARDWARE_SNK_ENCRYPTED_B)) {
		double fRange = (double)pInfo->nSpriteNum / 4.0;
		if (fRange < 1.5) {
			fRange = 1.5;
		}
		BurnSetProgressRange(1.0 / fRange);
	} else {
		BurnSetProgressRange(1.0 / pInfo->nSpriteNum);
	}

	// Load sprite data
	NeoLoadSprites(pInfo->nSpriteOffset, pInfo->nSpriteNum, NeoSpriteROM, nSpriteSize);

	NeoTextROM = (unsigned char*)malloc(nNeoTextROMSize + 0x020000);
	if (NeoTextROM == NULL) {
		return 1;
	}

	// Descramble C roms without messing out the text layers (matrimbl only)
	if (!strcmp(BurnDrvText(0), "matrimbl")) {
		cthd2003_c();
	}

	// Load Text layer tiledata
	{
		// Load boardROM data
		BurnLoadRom(NeoTextROM,	0x00080 + 0x14, 1);

		if (pInfo->nTextOffset != -1) {
			// Load S ROM data
			BurnLoadRom(NeoTextROM + 0x020000, pInfo->nTextOffset, 1);

		/* Original code for MAME by fataku */
		if (nNeoTextROMFixType == 1) {
		        unsigned char* srom = NeoTextROM+0x20000;
		        for (int i=0;i < 0x20000;i++) {
		        	srom[i]=BITSWAP08(srom[i],3,2,4,5,1,6,0,7);
		        }
		}
		/* Original code from mame traduced by FerchogtX */
		if (nNeoTextROMFixType == 2) {
			/* Descrambling S1 by FerchogtX, bassed on DorriGa's code */
			unsigned char* srom = NeoTextROM+0x20000;
			unsigned char* dst = (unsigned char*)malloc(0x20000);
			memcpy( dst+0x00000, srom+0x10000, 0x10000 );
			memcpy( dst+0x10000, srom+0x00000, 0x10000 );
			memcpy( srom, dst, 0x20000 );
			free(dst);
		}
		/* Thanks also to HappyASR for the info */
		if (nNeoTextROMFixType == 3) {
			unsigned char* srom = NeoTextROM+0x20000;
			unsigned char* dst = (unsigned char*)malloc(0x20000);
			if (dst)
			{
				memcpy(dst,srom,0x20000);
				// descrambling the S1 by dorriGa
				for (int j=0;j < 0x20000 ; j+=0x000010)
				{
					memcpy(srom+j, dst+j+0x000008,8);
					memcpy(srom+j+0x000008, dst+j,8);
				}
			}
			free(dst);
		}
		/* Thanks to Badzou for the info */
		if (nNeoTextROMFixType == 4) {
			/* Descrambling S1 by dorriga */
			unsigned char* srom = NeoTextROM+0x20000;
			unsigned char* dst = (unsigned char*)malloc(0x80000);
			memcpy( dst+0x00000, srom+0x60000, 0x20000 );
			memcpy( dst+0x20000, srom+0x40000, 0x20000 );
			memcpy( dst+0x40000, srom+0x20000, 0x20000 );
			memcpy( dst+0x60000, srom+0x00000, 0x20000 );
			memcpy( srom, dst, 0x80000 );
			free(dst);
		}
		// Converted by Jimmy_Page (www.neofighters.com)
		if (nNeoTextROMFixType == 5) {
		        unsigned char* srom = NeoTextROM+0x20000;
		        for (int i=0;i < 0x20000;i++) {
		        	srom[i]=BITSWAP08(srom[i],7,6,0,4,3,2,1,5);
		        }
		}
		// Original conversion by IQ_132 (This is for svcchaos, kof2003, and clones)
		if (nNeoTextROMFixType == 6) {
			unsigned char* srom = NeoTextROM+0x20000;
			for( int i=0;i < nNeoTextROMSize;i++ ){
				srom[i]=BITSWAP08(srom[i]^0xd2,4,0,7,2,5,1,6,3);
			}
		}

			// Save decrypted text ROM
			if (bsavedecrypteds1 == 1) {
				FILE* fp = fopen("xxx-s1d.bin", "wb");
				BurnUpdateProgress(0.0, "Saving decrypted S1 ROM...", 0);
				if (fp) {
					fwrite(NeoTextROM + 0x020000, 1, nNeoTextROMSize, fp);
					fclose(fp);
				}
			}

		}
		
	}

	// Decode text data
	BurnUpdateProgress(0.0, "Decoding text layer graphics...", 0);
	NeoDecodeText(NeoTextROM, nNeoTextROMSize);

	// Decode sprite data
	NeoDecodeSprites(NeoSpriteROM, nSpriteSize);
	
	if (bsavedecryptedcs == 1){
		FILE* fp = fopen("xxx-c1d.bin", "wb");
		BurnUpdateProgress(0.0, "Saving decrypted C ROM...", 0);
		if (fp) {
			fwrite(NeoSpriteROM, 1, nSpriteSize, fp);
			fclose(fp);
		}
	}

	// Allocate all memory is needed for ROM
	{
		int nLen;

		ROMIndex();													// Get amount of memory needed
		nLen = ROMEnd - (unsigned char*)0;
		if ((AllROM = (unsigned char*)malloc(nLen)) == NULL) {		// Allocate memory
			return 1;
		}
		memset(AllROM, 0, nLen);									// Initialise memory
		ROMIndex();													// Index the allocated memory
	}

	// Load the roms into memory
	if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SMA_PROTECTION) {
		BurnLoadRom(Neo68KROM01 + 0x0C0000, 0, 1);
		NeoLoadCode(pInfo->nCodeOffset + 1, pInfo->nCodeNum - 1, Neo68KROM02);
	} else {
		NeoLoadCode(pInfo->nCodeOffset, pInfo->nCodeNum, Neo68KROM01);
	}

	BurnLoadRom(NeoZ80ROM, pInfo->nSoundOffset, 1);

	if (pNeoInitCallback) {
		pNeoInitCallback();
	}

	// This is for saving the P roms
	if (bsavedecryptedps == 1)
	{
		FILE* PROM = fopen("xxx-p1d.bin", "wb");
		BurnUpdateProgress(0.0, "Saving descrambled P ROM...", 0);
		if (PROM) {
			fwrite(Neo68KROM01, 1, nCodeSize, PROM);
			fclose(PROM);
		}
	}

	if (NeoLoad68KBIOS((NeoSystem & 0x07) ^ 4)) {
		return 1;
	}

	BurnLoadRom(NeoZ80BIOS,	0x00080 + 0x10, 1);
	BurnLoadRom(NeoZoomROM,	0x00080 + 0x18, 1);

	if (pInfo->nADPCMANum) {
		char* pName;
		struct BurnRomInfo ri;
		unsigned char* pADPCMData;

		YM2610ADPCMAROM	= (unsigned char*)malloc(nYM2610ADPCMASize);
		if (YM2610ADPCMAROM == NULL) {
			return 1;
		}

		ri.nType = 0;
		ri.nLen = 0;
		BurnDrvGetRomInfo(&ri, pInfo->nADPCMOffset);
		BurnDrvGetRomName(&pName, pInfo->nADPCMOffset, 0);

		pADPCMData = YM2610ADPCMAROM;

		// pbobblen needs this (V ROMs are v3 & v4), note aof/wh1/wh1h (V ROMs are v2 & v4)
		if (pInfo->nADPCMANum == 2 && pName[FindType(pName) + 1] == '3') {
			pADPCMData += ri.nLen * 2;
		}

		NeoLoadADPCM(pInfo->nADPCMOffset, pInfo->nADPCMANum, pADPCMData);

		if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SWAPV) {
			for (int i = 0; i < 0x00200000; i++) {
				unsigned char n = YM2610ADPCMAROM[i];
				YM2610ADPCMAROM[i] = YM2610ADPCMAROM[0x00200000 + i];
				YM2610ADPCMAROM[0x00200000 + i] = n;
			}
		}
	}

	/* Neo-PCM2 PLAYMORE(c) 2002 Sound Chip Emulation */
	// some code by Fataku & some by the vconv (from 2ch) - FBA conversion by iq_132
	if (nNeoPCM2PLAYMOREByteswap != 0)
	{
		unsigned int addrs[7][2]={
			{0x000000,0xA5000}, //kof2002
			{0xFFCE20,0x01000}, //matrimelee
			{0xFE2CF6,0x4E001}, //mslug5
			{0xFEB2C0,0x0A000}, //samsho5
			{0xFFAC28,0xC2000}, //svcchaos
			{0xFF14EA,0xA7001}, //kof2003
			{0xFFB440,0x02000}, //samsh5sp
		};
		unsigned int xordata[7][8]={
			{0xF9,0xE0,0x5D,0xF3,0xEA,0x92,0xBE,0xEF}, //kof2002
			{0xC4,0x83,0xA8,0x5F,0x21,0x27,0x64,0xAF}, //matrimelee
			{0xC3,0xFD,0x81,0xAC,0x6D,0xE7,0xBF,0x9E}, //mslug5
			{0xCB,0x29,0x7D,0x43,0xD2,0x3A,0xC2,0xB4}, //samsho5
			{0xC3,0xFD,0x81,0xAC,0x6D,0xE7,0xBF,0x9E}, //svcchaos
			{0x4B,0xA4,0x63,0x46,0xF0,0x91,0xEA,0x62}, //kof2003
			{0x4B,0xA4,0x63,0x46,0xF0,0x91,0xEA,0x62}, //samsh5sp
		};

		int value = (nNeoPCM2PLAYMOREByteswap-1);	
		UINT8 *rom = YM2610ADPCMAROM;
		UINT8 *buf = (UINT8*)malloc(0x1000000);
		int ofst;

		memcpy(buf,rom,0x1000000);

		for(int i=0; i<0x1000000; i++)
		{
			ofst = (i & 0xfefffe) | ((i & 0x010000) >> 16) | ((i & 0x000001) << 16);
			ofst ^= addrs[value][1];
			rom[ofst] = (buf[((i + addrs[value][0]) & 0xffffff)] ^ xordata[value][(ofst & 0x07)]);
		}

		free(buf);
	}
	// Save decrypted V Roms
	if ((bsavedecryptedvs == 1) && (nNeoPCM2SNKByteswap != 0)) {
		FILE* VROM = fopen("xxx-v1d.bin", "wb");
		BurnUpdateProgress(0.0, "Saving decrypted V ROM...", 0);
		if (VROM) {
			fwrite(YM2610ADPCMAROM, 1, nYM2610ADPCMASize, VROM);
			fclose(VROM);
		}
	}

	/* Thanks to fataku and IQ_132 for the info */
	/* This is for matrim and clones */
	if (nADPCMByteswap == 1) {
		/* Here only 2 sample ROMs have byteswap,
		so we divide the function in two processes */
		BurnByteswap(YM2610ADPCMAROM+0x400000, 0x400000);
		BurnByteswap(YM2610ADPCMAROM+0xc00000, 0x400000);
	}
	/* This is for svcboot and clones */
	if (nADPCMByteswap == 2) {
		// svcboot/svcboota
		BurnLoadRom(YM2610ADPCMAROM+0x000000, 12, 1); // V2->V1
		BurnLoadRom(YM2610ADPCMAROM+0x400000, 11, 1); // V1->V2
		BurnLoadRom(YM2610ADPCMAROM+0x800000, 14, 1); // V4->V3
		BurnLoadRom(YM2610ADPCMAROM+0xc00000, 13, 1); // V3->V4
		BurnByteswap(YM2610ADPCMAROM, 0x1000000);
	}
	if (nADPCMByteswap == 3) {
		// svcplusa/svcplusb/svcsplus
		BurnLoadRom(YM2610ADPCMAROM+0x000000, 13, 1); // V2->V1
		BurnLoadRom(YM2610ADPCMAROM+0x400000, 12, 1); // V1->V2
		BurnLoadRom(YM2610ADPCMAROM+0x800000, 15, 1); // V4->V3
		BurnLoadRom(YM2610ADPCMAROM+0xc00000, 14, 1); // V3->V4
		BurnByteswap(YM2610ADPCMAROM, 0x1000000);
	}
	if (nADPCMByteswap == 4) {
		// svcplus
		BurnLoadRom(YM2610ADPCMAROM+0x000000, 14, 1); // V2->V1
		BurnLoadRom(YM2610ADPCMAROM+0x400000, 13, 1); // V1->V2
		BurnLoadRom(YM2610ADPCMAROM+0x800000, 16, 1); // V4->V3
		BurnLoadRom(YM2610ADPCMAROM+0xc00000, 15, 1); // V3->V4
		BurnByteswap(YM2610ADPCMAROM, 0x1000000);
	}
	// Save decrypted V Roms
	if ((bsavedecryptedvs == 1) && (nADPCMByteswap != 0)) {
		FILE* VROM = fopen("xxx-v1d.bin", "wb");
		BurnUpdateProgress(0.0, "Saving decrypted V ROM...", 0);
		if (VROM) {
			fwrite(YM2610ADPCMAROM, 1, nYM2610ADPCMASize, VROM);
			fclose(VROM);
		}
	}

	if (pInfo->nADPCMBNum) {
		YM2610ADPCMBROM	= (unsigned char*)malloc(nYM2610ADPCMBSize);
		if (YM2610ADPCMBROM == NULL) {
			return 1;
		}

		NeoLoadADPCM(pInfo->nADPCMOffset + pInfo->nADPCMANum, pInfo->nADPCMBNum, YM2610ADPCMBROM);
	} else {
		YM2610ADPCMBROM = YM2610ADPCMAROM;
		nYM2610ADPCMBSize = nYM2610ADPCMASize;
	}

	return 0;
}

// ----------------------------------------------------------------------------
// Bankswitch / memory map functions

static void NeoZ80SetBank0(int nBank)
{
	nBank &= 0x0F;
	if (nBank != nZ80Bank0) {
		unsigned char* nStartAddress = NeoZ80ROM + (nBank << 14);
		ZetMapArea(0x8000, 0xBFFF, 0, nStartAddress);
		ZetMapArea(0x8000, 0xBFFF, 2, nStartAddress);

		nZ80Bank0 = nBank;
	}

	return;
}

static void NeoZ80SetBank1(int nBank)
{
	nBank &= 0x1F;
	if (nBank != nZ80Bank1) {
		unsigned char* nStartAddress = NeoZ80ROM + (nBank << 13);
		ZetMapArea(0xC000, 0xDFFF, 0, nStartAddress);
		ZetMapArea(0xC000, 0xDFFF, 2, nStartAddress);

		nZ80Bank1 = nBank;
	}

	return;
}

static void NeoZ80SetBank2(int nBank)
{
	nBank &= 0x3F;
	if (nBank != nZ80Bank2) {
		unsigned char* nStartAddress = NeoZ80ROM + (nBank << 12);
		ZetMapArea(0xE000, 0xEFFF, 0, nStartAddress);
		ZetMapArea(0xE000, 0xEFFF, 2, nStartAddress);

		nZ80Bank2 = nBank;
	}

	return;
}

static void NeoZ80SetBank3(int nBank)
{
	nBank &= 0x7F;
	if (nBank != nZ80Bank3) {
		unsigned char* nStartAddress = NeoZ80ROM + (nBank << 11);
		ZetMapArea(0xF000, 0xF7FF, 0, nStartAddress);
		ZetMapArea(0xF000, 0xF7FF, 2, nStartAddress);

		nZ80Bank3 = nBank;
	}

	return;
}

static void NeoZ80MapROM(bool bMapBoardROM)
{
	if (bMapBoardROM) {
		// Bank in the Z80 boardROM
		ZetMapArea(0x0000, 0x7FFF, 0, NeoZ80BIOS);
		ZetMapArea(0x0000, 0x7FFF, 2, NeoZ80BIOS);
	} else {
		// Bank in the Z80 cartridge ROM
		ZetMapArea(0x0000, 0x7FFF, 0, NeoZ80ROM);
		ZetMapArea(0x0000, 0x7FFF, 2, NeoZ80ROM);
	}
}

static void MapVectorTable(bool bMapBoardROM)
{
	if (bMapBoardROM) {
		SekMapMemory(Neo68KVectors,	0x000000, 0x0003FF, SM_ROM);
	} else {
		SekMapMemory(Neo68KROM01, 0x000000, 0x0003FF, SM_ROM);
	}
}

inline static void MapPalette(int nBank)
{
	if (nNeoPaletteBank != nBank) {
		nNeoPaletteBank = nBank;
		SekMapMemory(NeoPalSrc[nBank], 0x400000, 0x401FFF, SM_ROM);

		NeoSetPalette();
	}
}

inline static void Bankswitch(unsigned int nBank)
{
	nBank &= 7;
	if (nBank >= (nCodeSize - 0x100000) >> 20) {
		nBank = 0;
	}

	if (nBank != nNeo68KROMBank) {
//		printf("Bankswitched main ROM, new bank is %i.\n", nBank);
		nNeo68KROMBank = nBank;
		SekMapMemory(Neo68KROM02 + nBank * 0x100000, 0x200000, 0x2FFFFF, SM_ROM);
	}
}

static void MapSMABank()
{
	SekMapMemory(Neo68KROM02 + nNeo68KROMBank, 0x200000, 0x2FE3FF, SM_ROM);
	if (nNeoSMARNGAddress[0] > 0 || nNeoSMARNGAddress[1] > 0) {
		SekMapMemory(Neo68KROM02 + nNeo68KROMBank + 0x0FE800, 0x2FE800, 0x2FFBFF, SM_ROM);
	} else {
		SekMapMemory(Neo68KROM02 + nNeo68KROMBank + 0x0FE800, 0x2FE800, 0x2FFFFF, SM_ROM);
	}
}

// -----------------------------------------------------------------------------
// Savestate support

int NeoScan(int nAction, int* pnMin)
{
	int nOldBIOS = nBIOS;
	struct BurnArea ba;

	if (pnMin) {												// Return minimum compatible version
		*pnMin = 0x029498;
	}

	// Make sure we have the correct value for nBIOS
	if (nAction & ACB_VOLATILE) {
		SCAN_VAR(nBIOS);
	}

	if (nAction & ACB_MEMCARD) {								// Scan memory card

		if (pnMin && (nAction & ACB_TYPEMASK) == ACB_MEMCARD) {	// Return minimum compatible version
			*pnMin = 0x029402;
		}

		ba.Data		= NeoMemoryCard;
		ba.nLen		= 0x004000;
		ba.nAddress = 0;
		ba.szName	= "Memory card";

		if ((nAction & ACB_TYPEMASK) == ACB_MEMCARD) {
			if (nAction & ACB_WRITE) {
				bMemoryCardInserted = true;
			}
			if (nAction & ACB_READ) {
				bMemoryCardInserted = false;

				// determine the size of the actual memory card inserted
				int nSize = (NeoMemoryCard[10] << 7) | (NeoMemoryCard[11] >> 1);
				if (nSize >= 0x0800) {
					ba.nLen = nSize;
				}
			}
		}

		BurnAcb(&ba);
	}
	if (!NeoCheckAESBIOS() && (nAction & ACB_NVRAM)) {			// Scan non-volatile memory

		if (pnMin && (nAction & ACB_TYPEMASK) == ACB_NVRAM) {	// Return minimum compatible version
			*pnMin = 0x029402;
		}

		ba.Data		= NeoNVRAM;
		ba.nLen		= 0x00010000;
		ba.nAddress = 0;
		ba.szName	= "NVRAM";
		BurnAcb(&ba);


		if ((BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK) == HARDWARE_SNK_GAMBLING) {
			ba.Data		= NeoNVRAM2;
			ba.nLen		= 0x00010000;
			ba.nAddress = 0;
			ba.szName	= "Extra NVRAM";
			BurnAcb(&ba);
		}

	}
	if (nAction & ACB_VOLATILE) {								// Scan all memory, devices & variables
		ba.Data		= Neo68KRAM;
		ba.nLen		= 0x00010000;
		ba.nAddress = 0;
		ba.szName	= "68K RAM";
		BurnAcb(&ba);

    	ba.Data		= NeoZ80RAM;
		ba.nLen		= 0x00000800;
		ba.nAddress = 0;
		ba.szName	= "Z80 RAM";
		BurnAcb(&ba);

    	ba.Data		= NeoPalSrc[0];
		ba.nLen		= 0x000002000;
		ba.nAddress = 0;
		ba.szName	= "Palette 0";
		BurnAcb(&ba);
    	ba.Data		= NeoPalSrc[1];
		ba.nLen		= 0x000002000;
		ba.nAddress = 0;
		ba.szName	= "Palette 1";
		BurnAcb(&ba);

    	ba.Data		= NeoGraphicsRAM;
		ba.nLen		= 0x00020000;
		ba.nAddress = 0;
		ba.szName	= "Graphics RAM";
		BurnAcb(&ba);

		SekScan(nAction);										// Scan 68000 state

		ZetOpen(0);
		ZetScan(nAction);										// Scan Z80 state
		ZetClose();

		BurnYM2610Scan(nAction, pnMin);
		BurnFMTimerScan(nAction);

		if (!NeoCheckAESBIOS()) {
			uPD4990AScan(nAction, pnMin);
		}

		SCAN_VAR(nCyclesDone);

		SCAN_VAR(bNeoEnableGraphics);
		SCAN_VAR(nIRQAcknowledge);

		SCAN_VAR(nIRQ2Control);
		SCAN_VAR(bSRAMWritable);

		SCAN_VAR(b68KBoardROMBankedIn);
		SCAN_VAR(bBIOSTextROMEnabled);
		SCAN_VAR(nZ80Bank0); SCAN_VAR(nZ80Bank1); SCAN_VAR(nZ80Bank2); SCAN_VAR(nZ80Bank3);
		SCAN_VAR(nNeo68KROMBank);

		NeoGraphicsRAMBank -= (unsigned int)NeoGraphicsRAM;
		SCAN_VAR(NeoGraphicsRAMBank); SCAN_VAR(NeoGraphicsRAMPointer); SCAN_VAR(nNeoGraphicsModulo);
		NeoGraphicsRAMBank += (unsigned int)NeoGraphicsRAM;

		SCAN_VAR(nNeoSpriteFrame); SCAN_VAR(nSpriteFrameSpeed); SCAN_VAR(nSpriteFrameTimer);

		SCAN_VAR(nNeoPaletteBank);

		SCAN_VAR(nSoundLatch);
		SCAN_VAR(nSoundReply); SCAN_VAR(nSoundPrevReply);
		SCAN_VAR(nSoundStatus);

		SCAN_VAR(nInputSelect);

		NeoInputBank -= (unsigned int)NeoInput;
		SCAN_VAR(NeoInputBank);
		NeoInputBank += (unsigned int)NeoInput;

		SCAN_VAR(nAnalogAxis);

		SCAN_VAR(nuPD4990ATicks);

		SCAN_VAR(nIRQLine); SCAN_VAR(nIRQLineLo); SCAN_VAR(nIRQLineHi);

		SCAN_VAR(nSMARandomNumber);

		if (nAction & ACB_WRITE) {
			int nNewBIOS = nBIOS;
			int nBank;

			MapVectorTable(b68KBoardROMBankedIn);
			NeoZ80MapROM(bZ80BoardROMBankedIn);

			nBank = nZ80Bank0; nZ80Bank0 = -1;
			NeoZ80SetBank0(nBank);
			nBank = nZ80Bank1; nZ80Bank1 = -1;
			NeoZ80SetBank1(nBank);
			nBank = nZ80Bank2; nZ80Bank2 = -1;
			NeoZ80SetBank2(nBank);
			nBank = nZ80Bank3; nZ80Bank3 = -1;
			NeoZ80SetBank3(nBank);

			if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SMA_PROTECTION || (ncthd2003==1) ) {
				MapSMABank();
			} else {
				nBank = nNeo68KROMBank; nNeo68KROMBank = -1U;
				Bankswitch(nBank);
			}

			nBank = nNeoPaletteBank; nNeoPaletteBank = -1;
			MapPalette(nBank);

			NeoRecalcPalette = 1;

			nBIOS = nOldBIOS;
			NeoLoad68KBIOS(nNewBIOS);

			bAESBIOS = NeoCheckAESBIOS();
		}
	}

	return 0;
}

// -----------------------------------------------------------------------------
// Z80 handlers

static void neogeoFMIRQHandler(int, int nStatus)
{
	if (nStatus) {
		nTriggerIRQ = 1;
	} else {
		ZetLowerIrq();
		nTriggerIRQ = -1;
	}
}

unsigned char __fastcall neogeoZ80In(unsigned short nAddress)
{
	switch (nAddress & 0xFF) {
		case 0x00:									// Read sound command
//			printf("  - Sound command received (0x%02X).\n", nSoundLatch);
			nSoundStatus = 1;
			nSoundPrevReply = -1;
			return nSoundLatch;

		case 0x04:
			return BurnYM2610Read(0);
		case 0x05:
			return BurnYM2610Read(1);
		case 0x06:
			return BurnYM2610Read(2);

		case 0x08:
//			printf("  - Z80 bank 3 -> 0x%02X\n", nAddress >> 8);
			NeoZ80SetBank3(nAddress >> 8);
			break;
		case 0x09:
//			printf("  - Z80 bank 2 -> 0x%02X\n", nAddress >> 8);
			NeoZ80SetBank2(nAddress >> 8);
			break;
		case 0x0A:
//			printf("  - Z80 bank 1 -> 0x%02X\n", nAddress >> 8);
			NeoZ80SetBank1(nAddress >> 8);
			break;
		case 0x0B:
//			printf("  - Z80 bank 0 -> 0x%02X\n", nAddress >> 8);
			NeoZ80SetBank0(nAddress >> 8);
			break;

//		default: {
//			printf("  - Z80 read port 0x%04X.\n", nAddress);
//		}
	}

	return 0;
}

void __fastcall neogeoZ80Out(unsigned short nAddress, unsigned char nValue)
{
	switch (nAddress & 0x0FF) {
		case 0x00:
//			printf("  - Z80 port 0x00 -> 0x%02X.\n", nValue);
			break;
		case 0x04:
		case 0x05:
		case 0x06:
		case 0x07:
			if (BurnYM2610Write(nAddress & 3, nValue)) {
//				ZetRaiseIrq(0xFF);
			}
			break;

		case 0x0C:									// Write reply to sound commands
//			printf("  - Sound reply sent (0x%02X).\n", nValue);
			nSoundReply = nValue;

			if (nSoundPrevReply != nValue) {
				nSoundPrevReply = nValue;

				// s1945p replies a 0x00, then an 0xFF;
				// the 68K loops until it has read both
				if (nSoundReply == 0) {
					nSoundStatus &= ~2;
				} else {
					nSoundStatus |=  2;
				}
			} else {
				nSoundStatus |= 2;
			}

			if (nCurrentCPU == 0) {
				ZetRunEnd();
			}

			break;

//		default: {
//			printf("  - Z80 port 0x%04X -> 0x%02X.\n", nAddress, nValue);
//		}
	}
}

// -----------------------------------------------------------------------------
// 68K handlers

static inline void NeoIRQAcknowledge(unsigned short wordValue)
{
	nIRQAcknowledge |= (wordValue & 6);

	if (nIRQAcknowledge == 6) {
		SekSetIRQLine(7, 0);
	} else {
		if ((nIRQAcknowledge & 2) == 0) {
			SekSetIRQLine(2, 1);
		}
		if ((nIRQAcknowledge & 4) == 0) {
			SekSetIRQLine(1, 1);
		}
	}

//	printf("  - IRQ Ack -> %02X   (at line %3i).\n", wordValue, NeoGetCurrentScanline());
}

static inline void SendSoundCommand(const char nCommand)
{
//	printf("  - Sound command sent (0x%02X).\n", nCommand);

	nCyclesDone[1] += ZetRun(((nCyclesDone[0] + SekMidCycles()) * 16384 / nNeoCPUSpeedfactor) - nCyclesDone[1]);

	nSoundStatus &= ~1;
	nSoundLatch = nCommand;

	ZetNmi();

	nCyclesDone[1] += ZetRun(0x0400);
}

static inline int NeoConvertIRQPosition()
{
	return ((nIRQLineHi << 16) | nIRQLineLo) / 384;
}

static inline int NeoConvertIRQPositionRelative()
{
	unsigned int nPos = ((nIRQLineHi << 16) | nIRQLineLo);
	nPos += nPos >> 4;

	return (nPos <= 384) ? 1 : (nPos / 384);
}

static inline unsigned short NeoGetCurrentScanline()
{
	return (nCyclesDone[0] + SekMidCycles()) / nCyclesScanline;
}

unsigned char __fastcall neogeoReadByte(unsigned int sekAddress)
{
	switch (sekAddress) {

		case 0x300000:
//			printf(" -- bank %d inputP0[0x%02X] read.\n", 0, nInputSelect);
			return ~NeoInputBank[nJoyport0[nInputSelect & 0x07]];

		case 0x300001:
			return ~NeoInputBank[4];

		case 0x300081:
			return  NeoInputBank[5];

		case 0x320000: {
			int nReply = nSoundReply;

			// nSoundStatus: &1 = sound latch read, &2 = response written
			if (nSoundStatus != 3) {
				nCyclesDone[1] += ZetRun(((nCyclesDone[0] + SekMidCycles()) * 16384 / nNeoCPUSpeedfactor) - nCyclesDone[1] + 0x0200);
			}

			if (nSoundStatus & 1) {
//				printf("  - Sound reply read (0x%02X).\n",  nSoundReply);

				return nReply;
			} else {
//				printf("  - Sound reply read while sound pending (0x%02X).\n",  nSoundReply);

				return nReply & 0x7F;
			}
		}

		case 0x320001: {
			if (!bAESBIOS) {
				unsigned char nuPD4990AOutput = uPD4990ARead(nCyclesDone[0] + SekMidCycles() - nuPD4990ATicks);
				nuPD4990ATicks = nCyclesDone[0] + SekMidCycles();
				return (~NeoInputBank[3] & 0x3F) | (nuPD4990AOutput << 6);
			}

			return (~NeoInputBank[3] & 0x7F);
		}

		case 0x340000:
			return ~NeoInputBank[nJoyport1[nInputSelect & 0x07]];

		case 0x380000:
//			printf(" -- input 2 read.\n");
			return ~NeoInputBank[2];

//		default:
//			printf("Attempt to read byte value of location %08X (PC: %08X)\n", sekAddress, SekPc(-1));
	}

	return 0;
}

unsigned short __fastcall neogeoReadWord(unsigned int sekAddress)
{
	switch (sekAddress) {
		case 0x3C0000:
		case 0x3C0002:
		case 0x3C000A:
//			printf("  - Graphics RAM read (Bank %i, address 0x%04X).\n", NeoGraphicsRAMPointer > 0xFFFF ? 1 : 0, NeoGraphicsRAMPointer & 0xFFFF);
			return *((unsigned short*)(NeoGraphicsRAMBank + NeoGraphicsRAMPointer));

		case 0x3C0004:
//			printf("  - Graphics RAM modulo read.\n");
			return (unsigned short)(nNeoGraphicsModulo >> 1);

		case 0x3C0006: {									// Display status
//			printf("  - Display status read, line: %3i, anim: %i\n", NeoGetCurrentScanline(), nNeoSpriteFrame);

			bForcePartialRender = bForceUpdateOnStatusRead;

			return ((NeoGetCurrentScanline() + nScanlineOffset) << 7) | (nNeoSpriteFrame & 7);
		}

//		default:
//			printf("Attempt to read word value of location %08X (PC: %08X)\n", sekAddress, SekPc(-1));
	}

	return 0;
}

void __fastcall neogeoWriteByte(unsigned int sekAddress, unsigned char byteValue)
{
	switch (sekAddress) {
		case 0x280051:										// Send command to RTC
		case 0x380051: {
			if (!bAESBIOS) {
				uPD4990AWrite(byteValue & 2, byteValue & 4, byteValue & 1);
			}
			break;
		}

		case 0x300001:										// Watchdog
//			printf("  - Watchdog timer reset (%02X, at scanline %i)\n", byteValue, NeoGetCurrentScanline());
			break;

		case 0x320000:
			SendSoundCommand(byteValue);
			break;

		case 0x380001: {									// Select the input returned at 0x300000
//			printf("  - InputP0/P1 0x%02X selected.\n", byteValue);
			nInputSelect = byteValue;
			break;
		}

		case 0x380031:										// bit 3 -> 0: reset Z80, bit 4/5 ???
//			printf("  - 0x%06X -> 0x%02X.\n", sekAddress, byteValue);
			if ((byteValue & 0x08) == 0) {

#ifdef Z80_SWITCH_ROMBANK_ON_RESET
				bZ80BoardROMBankedIn = bBIOSTextROMEnabled;
				NeoZ80MapROM(bZ80BoardROMBankedIn);
#endif

				nSoundStatus |= 1;
				ZetReset();
//				printf(" -- Z80 reset (%s ROM banked in).\n", bZ80BoardROMBankedIn ? "BIOS" : "cartridge");
			}
			break;
		case 0x380041:										// LEDs (bit 1: P1)
//			printf("  - LED output -> 0x%02X.\n", byteValue);
			break;

#if 0
		case 0x380061:										// ??? Coin chute 1 coin inserted (0x00 written after 0x3800E1)
			break;
		case 0x380063:										// ??? Coin chute 2 coin inserted (0x00 written after 0x3800E3)
			break;

		case 0x3800E1:										// bit 4: Coin chute 1 coin inserted
			break;
		case 0x3800E3:										// bit 4: Coin chute 2 coin inserted
			break;
#else
		case 0x380061:
//			printf("  - Input bank 0 selected.\n");
			NeoInputBank = NeoInput + 0;
			break;
		case 0x3800E1:
//			printf("  - Input bank 1 selected.\n");
			NeoInputBank = NeoInput + 8;
			break;
#endif

		case 0x3A0001: {									// Enable display
			bNeoEnableGraphics = true;
//			printf("  - Display  enabled (0x%02X, at scanline %i).\n", byteValue, NeoGetCurrentScanline());
			break;
		}
		case 0x3A0011: {									// Disable display
			bNeoEnableGraphics = false;
//			printf("  - Display disabled (0x%02X, at scanline %i).\n", byteValue, NeoGetCurrentScanline());
			break;
		}

		case 0x3A0003: {									// Select BIOS vector table
			if (!b68KBoardROMBankedIn) {
				MapVectorTable(true);
				b68KBoardROMBankedIn = true;
			}
//			printf("  - BIOS vector table banked in (0x%02X).\n", byteValue);
			break;
		}
		case 0x3A0013: {									// Select game vector table
			if (b68KBoardROMBankedIn) {
				MapVectorTable(false);
				b68KBoardROMBankedIn = false;
			}
//			printf("  - ROM vector table banked in (0x%02X).\n", byteValue);
			break;
		}

		case 0x3A000B: {									// Select BIOS text ROM
//			printf("  - BIOS text ROM banked in (0x%02X).\n", byteValue);

#ifdef Z80_SWITCH_ROMBANK_ON_RESET
			if (!bBIOSTextROMEnabled) {
				bBIOSTextROMEnabled = true;
				NeoZ80MapROM(true);

				nSoundStatus |= 1;
				ZetReset();
			}
#else
			bBIOSTextROMEnabled = true;
#endif
			break;
		}
		case 0x3A001B: {									// Select game text ROM
//			printf("  - Cartridge text ROM banked in (0x%02X).\n", byteValue);

#ifdef Z80_SWITCH_ROMBANK_ON_RESET
			if (bBIOSTextROMEnabled) {
				bBIOSTextROMEnabled = false;
				NeoZ80MapROM(false);

				nSoundStatus |= 1;
				ZetReset();
			}
#else
			bBIOSTextROMEnabled = false;
#endif
			break;
		}

		case 0x3A000D: {									// Write-protect SRAM
			bSRAMWritable = false;
//			printf("  - SRAM write-protected.\n");
			break;
		}
		case 0x3A001D: {									// Write-enable SRAM
			bSRAMWritable = true;
//			printf("  - SRAM writable.\n");
			break;
		}

		case 0x3C0006: {
			nSpriteFrameSpeed = byteValue;
//			printf("  - Sprite apeed -> 0x%02X\n", byteValue);
			break;
		}

		case 0x3A000F: {									// Select palette bank 1
//			printf("  - Palette 1 banked in.\n");
			MapPalette(1);
			break;
		}
		case 0x3A001F: {									// Select palette bank 0
//			printf("  - Palette 0 banked in.\n");
			MapPalette(0);
			break;
		}

//		case 0x3C0007:
//		case 0x3C000D: {
//			printf("  - Attempt to write byte 0x%06X ->   0x%02X\n", sekAddress, byteValue);
//			break;
//		}

//		default:
//			printf("  - Attempt to write byte 0x%06X ->   0x%02X\n", sekAddress, byteValue);
	}
}

void __fastcall neogeoWriteWord(unsigned int sekAddress, unsigned short wordValue)
{
	switch (sekAddress) {
		case 0x320000:
			SendSoundCommand(wordValue >> 8);
			break;

		case 0x3C0000: {
			NeoGraphicsRAMPointer = wordValue << 1;
			NeoGraphicsRAMBank = NeoGraphicsRAM;
			if (wordValue & 0x8000) {
				NeoGraphicsRAMBank += 0x00010000;
			}
			break;
		}
		case 0x3C0002: {
			*((unsigned short*)(NeoGraphicsRAMBank + NeoGraphicsRAMPointer)) = wordValue;
			NeoGraphicsRAMPointer += nNeoGraphicsModulo;
			break;
		}
		case 0x3C0004: {
			nNeoGraphicsModulo = ((short)wordValue) << 1;
			break;
		}

		case 0x3C0006: {
			nSpriteFrameSpeed = (wordValue >> 8);
#if 0
			if ((nIRQ2Control & 0x10) == 0 && wordValue & 0x10) {
				printf("  - IRQ2 enabled  (at line %3i, IRQControl: 0x%02X).\n", NeoGetCurrentScanline(), wordValue & 0xFF);
			} else {
				if (nIRQ2Control & 0x10 && (wordValue & 0x10) == 0) {
					printf("  - IRQ2 disabled (at line %3i, IRQControl: 0x%02X).\n", NeoGetCurrentScanline(), wordValue & 0xFF);
				}
			}
#endif

#if 0
			if ((nIRQ2Control & 0x10) == 0 && (wordValue & 0x10)) {
				NeoIRQAcknowledge(2);
			}
#endif
			if ((nIRQ2Control & 0x40) == 0 && (wordValue & 0x40)) {
				nIRQLine = NeoConvertIRQPosition() - nIRQOffset;
//				printf("  - IRQ Line -> %3i (at line %3i, ctrl: bit 6 set).\n", nIRQLine, NeoGetCurrentScanline());
			}

			nIRQ2Control = wordValue;
//			printf("  - Autoanim speed -> 0x%02X\n", wordValue >> 8);
//			printf("  - IRQ2 control register -> 0x%02X\n", wordValue & 0xFF);
			break;
		}

		case 0x3C0008: {
//			printf("0x%06X -> 0x%04X\n", sekAddress, wordValue);
			nIRQLineHi = wordValue;
			break;
		}
		case 0x3C000A: {
//			printf("0x%06X -> 0x%04X\n", sekAddress, wordValue);
			nIRQLineLo = wordValue;

//			printf("  - IRQ Pos  -> %3i (at line %3i).\n", NeoConvertIRQPosition(), NeoGetCurrentScanline());
			if (nIRQ2Control & 0x20) {
				if (NeoGetCurrentScanline() < 248) {
					nIRQLine = NeoGetCurrentScanline() + NeoConvertIRQPositionRelative();
//					printf("  - IRQ Line -> %3i (at line %3i, relative).\n", nIRQLine, NeoGetCurrentScanline());
				} else {
					nIRQLine = NeoConvertIRQPosition();
//					printf("  - IRQ Line -> %3i (at line %3i, ctrl: 0x%02X).\n", nIRQLine, NeoGetCurrentScanline(), nIRQ2Control & 0xFF);
					if (nIRQLine > 14 && nIRQLine < 264) {
						nIRQLine -= nIRQOffset;
					}
				}
			}
			break;
		}

		case 0x3C000C: {
			NeoIRQAcknowledge(wordValue);
			break;
		}

//		default:
//			printf("  - Attempt to write word 0x%06X -> 0x%04X\n", sekAddress, wordValue);
	}
}

// ----------------------------------------------------------------------------

void __fastcall neogeoWriteByteSRAM(unsigned int sekAddress, unsigned char byteValue)
{
	sekAddress &= 0xFFFF;

	if (bSRAMWritable) {
		if (byteValue == 1) {
			if ((int)(sekAddress ^ 1) != nSRAMProtAddress) {
				NeoNVRAM[sekAddress ^ 1] = byteValue;
			}
		} else {
			NeoNVRAM[sekAddress ^ 1] = byteValue;
		}
	}
}

void __fastcall neogeoWriteWordSRAM(unsigned int sekAddress, unsigned short wordValue)
{
	sekAddress &= 0xFFFF;

	if (bSRAMWritable) {
		if ((wordValue & 0xFF) == 1) {
			if ((int)sekAddress != nSRAMProtAddress) {
				*((unsigned short*)(NeoNVRAM + sekAddress)) = wordValue;
			} else {
				NeoNVRAM[sekAddress] = wordValue >> 8;
			}
		} else {
			*((unsigned short*)(NeoNVRAM + sekAddress)) = wordValue;
		}
	}
}

// ----------------------------------------------------------------------------

unsigned char __fastcall neogeoReadByteMemoryCard(unsigned int sekAddress)
{
//	printf("  - Memcard 0x%04X read (PC: 0x%06X).\n", sekAddress & 0x7FFF, SekPc(-1));

	if (bMemoryCardInserted && (sekAddress & 1) == 1) {
		sekAddress &= 0x7FFF;
		return NeoMemoryCard[sekAddress >> 1];
	}

	return 0xFF;
}

void __fastcall neogeoWriteByteMemoryCard(unsigned int sekAddress, unsigned char byteValue)
{
//	printf("  - Memcard 0x%04X -> 0x%02X (PC: 0x%06X).\n", sekAddress & 0x7FFF, byteValue, SekPc(-1));

	if (bMemoryCardInserted && (sekAddress & 1) == 1 && bMemoryCardWritable) {
		sekAddress &= 0x7FFF;
		NeoMemoryCard[sekAddress >> 1] = byteValue;
	}
}

// ----------------------------------------------------------------------------
// 68K bankswitch for games without SMA protection

void __fastcall neogeoWriteByteBankswitch(unsigned int sekAddress, unsigned char byteValue)
{
	switch (sekAddress) {
//		case 0x2FFFF0:
		case 0x2FFFF1:
//		case 0x2FFFF2:
		case 0x2FFFF3:
//		case 0x2FFFF4:
		case 0x2FFFF5:
//		case 0x2FFFF6:
		case 0x2FFFF7:
//		case 0x2FFFF8:
		case 0x2FFFF9:
//		case 0x2FFFFA:
		case 0x2FFFFB:
//		case 0x2FFFFC:
		case 0x2FFFFD:
//		case 0x2FFFFE:
		case 0x2FFFFF: {
			Bankswitch(byteValue);
		}
	}
}

void __fastcall neogeoWriteWordBankswitch(unsigned int sekAddress, unsigned short wordValue)
{
	switch (sekAddress) {
		case 0x2FFFF0:
		case 0x2FFFF1:
		case 0x2FFFF2:
		case 0x2FFFF3:
		case 0x2FFFF4:
		case 0x2FFFF5:
		case 0x2FFFF6:
		case 0x2FFFF7:
		case 0x2FFFF8:
		case 0x2FFFF9:
		case 0x2FFFFA:
		case 0x2FFFFB:
		case 0x2FFFFC:
		case 0x2FFFFD:
		case 0x2FFFFE:
		case 0x2FFFFF: {
			Bankswitch(wordValue);
		}
	}
}

// ----------------------------------------------------------------------------
// SMA chip protection and random number generator

unsigned short __fastcall neogeoReadWordSMA9A37(unsigned int sekAddress)
{
//	printf(" -- SMA9A37 0x%08X (word)\n", sekAddress);

	if (sekAddress == 0x2FE446) {
		return 0x9A37;
	}

	return *((unsigned short*)(Neo68KROM02 + nNeo68KROMBank + sekAddress - 0x200000));
}

unsigned char __fastcall neogeoReadByteSMA9A37(unsigned int sekAddress)
{
//	printf(" -- SMA9A37 0x%08X (byte)\n", sekAddress);

	if (sekAddress == 0x2FE446) {
		return 0x9A;
	}
	if (sekAddress == 0x2FE447) {
		return 0x37;
	}

	return Neo68KROM02[(nNeo68KROMBank + sekAddress - 0x200000) ^ 1];
}

unsigned short __fastcall neogeoReadWordSMARNG(unsigned int sekAddress)
{
	if (sekAddress == nNeoSMARNGAddress[0] || sekAddress == nNeoSMARNGAddress[1]) {
		int nRandomNubmer = nSMARandomNumber;

		int nNewbit = (
				(nSMARandomNumber >>  2) ^
				(nSMARandomNumber >>  3) ^
				(nSMARandomNumber >>  5) ^
				(nSMARandomNumber >>  6) ^
				(nSMARandomNumber >>  7) ^
				(nSMARandomNumber >> 11) ^
				(nSMARandomNumber >> 12) ^
				(nSMARandomNumber >> 15)) & 1;

		nSMARandomNumber = ((nSMARandomNumber << 1) | nNewbit) & 0x0FFFFF;

		return nRandomNubmer;
	}

	return *((unsigned short*)(Neo68KROM02 + nNeo68KROMBank + sekAddress - 0x200000));
}

unsigned char __fastcall neogeoReadByteSMARNG(unsigned int sekAddress)
{
	if ((sekAddress & ~1) == nNeoSMARNGAddress[0] || (sekAddress & ~1) == nNeoSMARNGAddress[1]) {
		int nRandomNubmer = nSMARandomNumber;

		int nNewbit = (
				(nSMARandomNumber >>  2) ^
				(nSMARandomNumber >>  3) ^
				(nSMARandomNumber >>  5) ^
				(nSMARandomNumber >>  6) ^
				(nSMARandomNumber >>  7) ^
				(nSMARandomNumber >> 11) ^
				(nSMARandomNumber >> 12) ^
				(nSMARandomNumber >> 15)) & 1;

		nSMARandomNumber = ((nSMARandomNumber << 1) | nNewbit) & 0x0FFFFF;

		if (sekAddress & 1) {
			return nRandomNubmer & 0xFF;
		} else {
			return nRandomNubmer >> 8;
		}
	}

	return Neo68KROM02[(nNeo68KROMBank + sekAddress - 0x200000) ^ 1];
}

// ----------------------------------------------------------------------------

static int neogeoReset()
{
	NeoLoad68KBIOS((NeoSystem & 0x07) ^ 4);

	if (nBIOS == -1 || nBIOS == 9) {
		// Write system type & region code into BIOS ROM
		*((unsigned short*)(Neo68KBIOS + 0x000400)) = ((NeoSystem & 4) << 13) | (NeoSystem & 0x03);
	}

#if 0 && defined _DEBUG
	printf("  - Emulating ");
	if (nBIOS < 8 || nBIOS == 9) {
		switch (NeoSystem & 0x03) {
			case 0x00:
				printf("Japanese");
				break;
			case 0x01:
				printf("US");
				break;
			case 0x02:
				printf("Asian / European");
				break;
			default:
				printf("unknown");
				break;
		}
		if (NeoSystem & 0x04) {
			printf(" MVS system");
		} else {
			printf(" AES system");
		}
	}
	if (nBIOS >= 8) {
		switch (nBIOS) {
			case 0x08:
				printf("development system");
				break;
			case 0x09:
				printf(" with trackball control");
				break;
			case 0x0C:
				printf("MVS system with Universe BIOS v1.2");
				break;
			default:
				printf("unknown system");
				break;
		}
	}
	if ((BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK) == HARDWARE_SNK_PADDLE) {
		printf(" with paddle support");
	}
	if (nBIOS == -1) {
		printf(" (patching region)");
	}
	if (nBIOS == 9 && NeoSystem != 6) {
		printf(" (patching region and/or system type)");
	}
	printf(".\n");
#endif

	bSRAMWritable = false;
	// Enable SRAM protection when emulating an MVS system
	nSRAMProtAddress = nNeoSRAMProtection;

	bNeoEnableGraphics = true;

	bBIOSTextROMEnabled = false;
	bZ80BoardROMBankedIn = false;
	b68KBoardROMBankedIn = true;

	nNeoPaletteBank = -1;

	nSpriteFrameSpeed = 4;
	nSpriteFrameTimer = 0;
	nNeoSpriteFrame = 0;

	nIRQAcknowledge = 6;
	bIRQ2Enabled = false;

	nSoundLatch = 0x00;
	nSoundReply = 0x00;
	nSoundPrevReply = -1;
	nSoundStatus = 1;

	nIRQLine = nIRQLineLo = nIRQLineHi = 0;
	nIRQ2Control = 0;

	nInputSelect = 0;
	NeoInputBank = NeoInput;

	nSMARandomNumber = 0x2345;

	nCyclesDone[0] = nCyclesDone[1] = 0;

	bAESBIOS = NeoCheckAESBIOS();

	{
		SekOpen(0);

		if (bAESBIOS) {
			SekMapHandler(0,			0xD00000, 0xD0FFFF, SM_RAM);	// AES doesn't have the SRAM
		} else {
			SekMapMemory(NeoNVRAM,		0xD00000, 0xD0FFFF, SM_ROM);	// Battery backed SRAM
			SekMapHandler(1,			0xD00000, 0xD0FFFF, SM_WRITE);	//
		}

		MapVectorTable(b68KBoardROMBankedIn);

		SekReset();

		MapPalette(0);

		if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SMA_PROTECTION || (ncthd2003==1) ) {
			nNeo68KROMBank = 0;
			MapSMABank();
		} else {
			if ((BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK) != HARDWARE_SNK_GAMBLING) {
					//// added
					if (!(BurnDrvGetHardwareCode() & HARDWARE_SNK_NEWBANKSYSTEM)) {
					////end
						nNeo68KROMBank = -1U;
						Bankswitch(0);
					//// added
					}
					////end
				}
		}

		SekClose();
	}

	{
		ZetOpen(0);

		NeoZ80MapROM(bZ80BoardROMBankedIn);

		nZ80Bank0 = nZ80Bank1 = nZ80Bank2 = nZ80Bank3 = -1;
		NeoZ80SetBank0(0x02);
		NeoZ80SetBank1(0x06);
		NeoZ80SetBank2(0x0E);
		NeoZ80SetBank3(0x1E);

		ZetReset();

		ZetClose();
	}

	BurnYM2610Reset();
	uPD4990AInit((int)((long long)12000000 * nBurnCPUSpeedAdjust / 256)); // This is used to avoid calendar errors while overclockng a neogeo game

	return 0;
}

int NeoInit()
{
	NeoGameInfo info;
	NeoGameInfo* pInfo = &info;

	{
		struct BurnRomInfo ri;

		ri.nType = 0;
		ri.nLen = 0;

		if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SRAM && nNeoSRAMProtection == -1) {
			nNeoSRAMProtection = 0x0100;
		}

		// Find 'P' ROMs
		FindROMs("p ", &pInfo->nCodeOffset, &pInfo->nCodeNum);
		// Find 'S' ROM
		FindROMs("s ", &pInfo->nTextOffset, NULL);
		// Find 'C' ROMs
		FindROMs("c ", &pInfo->nSpriteOffset, &pInfo->nSpriteNum);
		// Find 'M' ROM
		FindROMs("m ", &pInfo->nSoundOffset, NULL);
		// Find 'V' ROMs
		FindROMs("v ", &pInfo->nADPCMOffset, &pInfo->nADPCMANum);
		if (pInfo->nADPCMOffset >= 0) {
			FindROMs("v2", NULL, &pInfo->nADPCMBNum);
		}

		if (pInfo->nADPCMBNum < 0) {
			pInfo->nADPCMBNum = 0;
		}
		pInfo->nADPCMANum -= pInfo->nADPCMBNum;

#if 0 && defined _DEBUG
		printf("  - P: %i (%i);", pInfo->nCodeOffset, pInfo->nCodeNum);
		if (pInfo->nTextOffset >= 0) {
			printf(" S: %i;", pInfo->nTextOffset);
		} else {
			printf(" S: unused;");
		}
		printf(" C: %i (%i); M: %i", pInfo->nSpriteOffset, pInfo->nSpriteNum, pInfo->nSoundOffset);
		if (pInfo->nADPCMOffset >= 0) {
			printf(" V: %i (%i, %i)", pInfo->nADPCMOffset, pInfo->nADPCMANum, pInfo->nADPCMBNum);
		} else {
			printf(" V: unused");
		}
		printf("\n");
#endif

		nCodeSize = 0;
		for (int i = 0; i < pInfo->nCodeNum; i++) {
			BurnDrvGetRomInfo(&ri, pInfo->nCodeOffset + i);
			nCodeSize += ri.nLen;
		}
		nCodeSize = (nCodeSize + 0x0FFFFF) & ~0x0FFFFF;

		nSpriteSize = 0;

		// for viewpoin, aof, ssideki & kotm2
		if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SWAPC) {
			BurnDrvGetRomInfo(&ri, pInfo->nSpriteOffset);
			if (ri.nLen == 0x200000 && pInfo->nSpriteNum < 4) {
				nSpriteSize = 0x600000;
			}
		}

		if (nSpriteSize == 0) {

			// Compute correct size taking gaps into account (kizuna)
			for (int i = 0; i < pInfo->nSpriteNum - 2; i += 2) {
				BurnDrvGetRomInfo(&ri, pInfo->nSpriteOffset + i);
				if (ri.nLen > nSpriteSize) {
					nSpriteSize = ri.nLen;
				}
			}
			nSpriteSize *= pInfo->nSpriteNum - 2;

			// The final 2 ROMs may have a different size
			BurnDrvGetRomInfo(&ri, pInfo->nSpriteOffset + pInfo->nSpriteNum - 2);
			nSpriteSize += ri.nLen * 2;
		}

		for (nNeoTileMask = 1; nNeoTileMask < nSpriteSize; nNeoTileMask <<= 1) {}
		nNeoTileMask = (nNeoTileMask >> 7) - 1;
		nNeoMaxTile = nSpriteSize >> 7;

		if (nNeoTextROMSize == -1) {
			if (pInfo->nTextOffset >= 0) {
				BurnDrvGetRomInfo(&ri, pInfo->nTextOffset);
				nNeoTextROMSize = ri.nLen;
			} else {
				nNeoTextROMSize = 0x020000;
			}
		}

		if (!strcmp(BurnDrvText(0), "irrmaze")) {
			pInfo->nADPCMANum--;
			pInfo->nADPCMBNum++;
		}

		nYM2610ADPCMASize = nYM2610ADPCMBSize = 0;
		if (pInfo->nADPCMOffset >= 0)	{
			char* pName;
			BurnDrvGetRomInfo(&ri, pInfo->nADPCMOffset);
			BurnDrvGetRomName(&pName, pInfo->nADPCMOffset, 0);
			nYM2610ADPCMASize = ri.nLen;

			if (pInfo->nADPCMANum > 1) {
				BurnDrvGetRomInfo(&ri, pInfo->nADPCMOffset + pInfo->nADPCMANum - 1);
				BurnDrvGetRomName(&pName, pInfo->nADPCMOffset + pInfo->nADPCMANum - 1, 0);
				if (pInfo->nADPCMBNum == 0) {
					nYM2610ADPCMASize *= pName[5] - '1';
				} else {
					nYM2610ADPCMASize *= pName[6] - '1';
				}
				nYM2610ADPCMASize += ri.nLen;
			}

			if (pInfo->nADPCMBNum) {
				BurnDrvGetRomInfo(&ri, pInfo->nADPCMOffset + pInfo->nADPCMANum);
				nYM2610ADPCMBSize = ri.nLen * (pInfo->nADPCMBNum - 1);
				BurnDrvGetRomInfo(&ri, pInfo->nADPCMOffset + pInfo->nADPCMANum + pInfo->nADPCMBNum - 1);
				nYM2610ADPCMBSize += ri.nLen;
			}
		}
	}

	nBIOS = 9999;
	if (LoadRoms(pInfo)) {
		return 1;
	}

	// Allocate all memory is needed for RAM
	{
		int nLen;

		RAMIndex();													// Get amount of memory needed
		nLen = RAMEnd - (unsigned char*)0;
		if ((AllRAM = (unsigned char*)malloc(nLen)) == NULL) {		// Allocate memory
			return 1;
		}
		memset(AllRAM, 0, nLen);									// Initialise memory
		RAMIndex();													// Index the allocated memory
	}

	{
		SekInit(1);													// Allocate 68000

		// Map 68000 memory:
	    SekOpen(0);

		SekMapMemory(Neo68KROM01,	0x000000, 0x0FFFFF, SM_ROM);	// 68K ROM bank 1
		SekMapMemory(Neo68KRAM,		0x100000, 0x10FFFF, SM_RAM);	// 68K RAM
		SekMapMemory(NeoPalSrc[0],	0x6A0000, 0x6A1FFF, SM_RAM);	// Palette RAM bank 0 always accessed here
		SekMapMemory(Neo68KBIOS,	0xC00000, 0xC1FFFF, SM_ROM);	// BIOS ROM

		SekSetReadWordHandler(0, neogeoReadWord);
		SekSetReadByteHandler(0, neogeoReadByte);
		SekSetWriteWordHandler(0, neogeoWriteWord);
		SekSetWriteByteHandler(0, neogeoWriteByte);

		SekSetWriteWordHandler(1, neogeoWriteWordSRAM);
		SekSetWriteByteHandler(1, neogeoWriteByteSRAM);

		SekMapHandler(2,			0x800000, 0x807FFF, SM_ROM);	// Memory card
		SekMapHandler(2,			0x800000, 0x807FFF, SM_WRITE);	//

		SekSetReadByteHandler(2, neogeoReadByteMemoryCard);
		SekSetWriteByteHandler(2, neogeoWriteByteMemoryCard);

		SekMapHandler(3,	0x400000, 0x401FFF, SM_WRITE);			// Palette

		SekSetWriteWordHandler(3, NeoPalWriteWord);
		SekSetWriteByteHandler(3, NeoPalWriteByte);

		if (BurnDrvGetHardwareCode() & HARDWARE_SNK_SMA_PROTECTION || (ncthd2003==1) ) {
			SekMapHandler(4,		0x2FFC00, 0x2FFFFF, SM_WRITE);	// Bankswitch

			SekSetWriteWordHandler(4, pNeoSMABankswitch);

			SekMapHandler(5,		0x2FE400, 0x2FE7FF, SM_ROM);	// Protection

			SekSetReadWordHandler(5, neogeoReadWordSMA9A37);
			SekSetReadByteHandler(5, neogeoReadByteSMA9A37);

			if (nNeoSMARNGAddress[0] > 0 || nNeoSMARNGAddress[1] > 0) {
				SekMapHandler(6,	0x2FFC00, 0x2FFFFF, SM_ROM);	// Randum number generator

				SekSetReadWordHandler(6, neogeoReadWordSMARNG);
				SekSetReadByteHandler(6, neogeoReadByteSMARNG);
			}
		} else {
			if ((BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK) == HARDWARE_SNK_GAMBLING) {
				SekMapMemory(NeoNVRAM2,	0x200000, 0x201FFF, SM_RAM);	// 68K RAM
			} else {
				if (nCodeSize > 0x100000) {
					SekMapHandler(4,	0x200000, 0x2FFFFF, SM_WRITE);	// 68K bankswitch

					SekSetWriteWordHandler(4, neogeoWriteWordBankswitch);
					SekSetWriteByteHandler(4, neogeoWriteByteBankswitch);
				}
			}
		}
		
		SekClose();
	}

	{
		// Z80 setup
		ZetInit(1);
		ZetOpen(0);

		// Work RAM
		ZetMapArea(0xF800, 0xFFFF, 0, NeoZ80RAM);
		ZetMapArea(0xF800, 0xFFFF, 1, NeoZ80RAM);
		ZetMapArea(0xF800, 0xFFFF, 2, NeoZ80RAM);

		ZetMemEnd();

		ZetSetInHandler(neogeoZ80In);
		ZetSetOutHandler(neogeoZ80Out);

		ZetClose();
	}

	bRenderLineByLine = false;
	if (!strcmp(BurnDrvText(0), "mosyougi")) {
		bRenderLineByLine = true;
	}

	nIRQOffset = 14;
	nScanlineOffset = 0xF8;										// correct as verified on MVS hardware

#ifdef IRQ_KLUDGE
	{
		// Per-game kludges

		if (!strcmp(BurnDrvText(0), "aodk")) {
			nIRQOffset = 15;
		}
		if (!strcmp(BurnDrvText(0), "galaxyfg")) {
			nIRQOffset = 15;
		}
		if (!strcmp(BurnDrvText(0), "mosyougi")) {
			nScanlineOffset = 0xF9;
		}
		if (!strcmp(BurnDrvText(0), "neodrift")) {
			nIRQOffset = 9;
		}
		if (!strncmp(BurnDrvText(0), "wh1", 3)) {
			nIRQOffset = 15;
		}
		if (!strcmp(BurnDrvText(0), "wh2")) {
			nIRQOffset = 15;
		}
		if (!strcmp(BurnDrvText(0), "zedblade")) {
			nIRQOffset = 0;
			nScanlineOffset = 0xF7;
		}
	}
#endif

	nNeoControlConfig = BurnDrvGetHardwareCode() & HARDWARE_SNK_CONTROLMASK;

	for (int i = 0; i < 8; i++) {
		nJoyport0[i] = 0;
		nJoyport1[i] = 1;
	}

	switch (nNeoControlConfig) {
		case HARDWARE_SNK_PADDLE:									// Two Paddles + joysticks
			nJoyport0[0] = 6;
			nJoyport1[0] = 7;
			break;
		case HARDWARE_SNK_TRACKBALL:								// Trackball controller
			nJoyport0[0] = 6;
			nJoyport0[1] = 7;
			break;
		case HARDWARE_SNK_MAHJONG:									// Mahjong controller
			nJoyport0[1] = 16;
			nJoyport0[2] = 17;
			nJoyport0[4] = 18;
			break;
	}

	BurnYM2610Init(8000000, YM2610ADPCMAROM, &nYM2610ADPCMASize, YM2610ADPCMBROM, &nYM2610ADPCMBSize, &neogeoFMIRQHandler);

	NeoInitText();
	NeoInitSprites();

	NeoInitPalette();

	uPD4990AInit((int)((long long)12000000 * nBurnCPUSpeedAdjust / 256)); // This is used to avoid calendar errors while overclocking a neogeo game

	bMemoryCardInserted = false;
	bMemoryCardWritable = false;

	neogeoReset();							// Reset machine

	return 0;
}

int NeoExit()
{
	uPD4990AExit();

	NeoExitPalette();

	NeoExitSprites();
	NeoExitText();

	BurnYM2610Exit();

	ZetExit();								// Deallocate Z80
	SekExit();								// Deallocate 68000

	// Deallocate all used memory
	free(NeoTextROM);						// Text ROM
	NeoTextROM = NULL;
	free(NeoSpriteROM);						// Sprite ROM
	NeoSpriteROM = NULL;

	if (nYM2610ADPCMASize) {				// ADPCM data
		free(YM2610ADPCMAROM);
	}
	if (YM2610ADPCMBROM != YM2610ADPCMAROM) {
		free(YM2610ADPCMBROM);
	}
	YM2610ADPCMAROM = NULL;
	YM2610ADPCMBROM = NULL;

	Neo68KROM01 = NULL;						// 68000 ROM
	Neo68KROM02 = NULL;

	free(AllROM);							// Misc ROM
	AllROM = NULL;

	free(AllRAM);							// Misc RAM
	AllRAM = NULL;

	nNeoSMARNGAddress[0] = nNeoSMARNGAddress[1] = 0;

	nNeoSRAMProtection = -0;
	nNeoPCM2PLAYMOREByteswap = 0; // exit PCM2 PLAYMORE chip
	nNeoTextROMSize = -1;
	ncthd2003 =-1;
	nNeoPCM2SNKByteswap = 0; // Turns off the NEO-PCM2 code when you quit a game
	nADPCMByteswap = 0;
	nNeoTextROMFixType = 1; // Turns off Text fix when you quit a game

	bsavedecrypteds1 = 0;
	bsavedecryptedvs = 0;
	bsavedecryptedps = 0;
	bsavedecryptedcs = 0;

	nBIOS = 9999;

	pNeoInitCallback = NULL;

	return 0;
}

int NeoRender()
{
	NeoUpdatePalette();						// Update the palette
	NeoClearScreen();

	if (bNeoEnableGraphics) {
		nSliceStart = 0x10; nSliceEnd = 0xF0;
		nSliceSize = nSliceEnd - nSliceStart;
		NeoRenderSprites();					// Render sprites
		NeoRenderText();					// Render text layer
	}

	return 0;
}

inline static int CheckSleep(int)
{
	return 0;
}

inline static void NeoClearOpposites(unsigned char* nJoystickInputs)
{
	if ((*nJoystickInputs & 0x03) == 0x03) {
		*nJoystickInputs &= ~0x03;
	}
	if ((*nJoystickInputs & 0x0C) == 0x0C) {
		*nJoystickInputs &= ~0x0C;
	}
}

static void NeoStandardInputs(int nBank)
{
	if (nBank) {
		NeoInput[ 8] = 0x00;					   					// Player 1
		NeoInput[ 9] = 0x00;				   						// Player 2
		NeoInput[10] = 0x00;				   						// Buttons
		NeoInput[11] = 0x00;				   						//
		for (int i = 0; i < 8; i++) {
			NeoInput[ 8] |= (NeoJoy3[i] & 1) << i;
			NeoInput[ 9] |= (NeoJoy4[i] & 1) << i;
			NeoInput[10] |= (NeoButton3[i] & 1) << i;
			NeoInput[11] |= (NeoButton4[i] & 1) << i;
		}
		NeoClearOpposites(&NeoInput[ 8]);
		NeoClearOpposites(&NeoInput[ 9]);

		if (NeoDiag[1]) {
			NeoInput[13] &= ~0x80;
		}
	} else {
		NeoInput[ 0] = 0x00;					   					// Player 1
		NeoInput[ 1] = 0x00;					   					// Player 2
		NeoInput[ 2] = 0x00;					   					// Buttons
		NeoInput[ 3] = 0x00;					   					//
		for (int i = 0; i < 8; i++) {
			NeoInput[ 0] |= (NeoJoy1[i] & 1) << i;
			NeoInput[ 1] |= (NeoJoy2[i] & 1) << i;
			NeoInput[ 2] |= (NeoButton1[i] & 1) << i;
			NeoInput[ 3] |= (NeoButton2[i] & 1) << i;
		}
		NeoClearOpposites(&NeoInput[ 0]);
		NeoClearOpposites(&NeoInput[ 1]);

		if (NeoDiag[0]) {
			NeoInput[ 5] &= ~0x80;
		}
	}
}

int NeoFrame()
{
	const int nInterleave = 10;

	static bool bPrevIRQ2Enabled;

	int nSoundBufferPos = 0;
	double dT = dTime;

	nuPD4990ATicks = 0;

	if (NeoReset) {							   						// Reset machine
		neogeoReset();
	}

	NeoInput[ 5] |= 0x80;											// Default test switches to off
	NeoInput[13] |= 0x80;											//

	switch (nNeoControlConfig) {
		case HARDWARE_SNK_PADDLE: {									// Two Paddles + joysticks

			NeoStandardInputs(0);

			// Handle analog controls
			nAnalogAxis[0] -= NeoAxis[0];
			nAnalogAxis[1] -= NeoAxis[1];
			NeoInput[6] = (nAnalogAxis[0] >> 8) & 0xFF;
			NeoInput[7] = (nAnalogAxis[1] >> 8) & 0xFF;

			break;
		}

		case HARDWARE_SNK_TRACKBALL: {								// Trackball controller
			NeoInput[1] = 0x00;				   						// Buttons
			NeoInput[2] = 0x00;					   					//
			NeoInput[3] = 0x00;					   					//
			for (int i = 0; i < 8; i++) {
				NeoInput[1] |= (NeoJoy2[i] & 1) << i;
				NeoInput[2] |= (NeoButton1[i] & 1) << i;
				NeoInput[3] |= (NeoButton2[i] & 1) << i;
			}
			// Handle analog controls
			nAnalogAxis[0] += NeoAxis[0];
			nAnalogAxis[1] += NeoAxis[1];
			NeoInput[6] = (nAnalogAxis[0] >> 8) & 0xFF;
			NeoInput[7] = (nAnalogAxis[1] >> 8) & 0xFF;

			if (NeoDiag[0]) {
				NeoInput[5] &= ~0x80;
			}

			break;
		}

		case HARDWARE_SNK_4_JOYSTICKS: {							// Four joystick controllers

			NeoStandardInputs(0);
			NeoStandardInputs(1);

			break;
		}

		case HARDWARE_SNK_MAHJONG: {								// Mahjong controller

			NeoStandardInputs(0);

			NeoInput[16] = 0x00;
			NeoInput[17] = 0x00;
			NeoInput[18] = 0x00;
			for (int i = 0; i < 7; i++) {
				NeoInput[16] |= (NeoButton1[i +  8] & 1) << i;
				NeoInput[17] |= (NeoButton1[i + 16] & 1) << i;
				NeoInput[18] |= (NeoButton1[i + 24] & 1) << i;
			}

			break;
		}

		case HARDWARE_SNK_GAMBLING: {								// Gambling configuration

			NeoStandardInputs(0);
			// Special inputs

			break;
		}

		default: {													// Two joystick controllers
			NeoStandardInputs(0);
		}
	}

	bMemoryCardWritable = (NeoSystem & 0x80);

	if (bMemoryCardInserted) {
		NeoInput[2] |= 0xB0;
		if (bMemoryCardWritable) {
			NeoInput[2] |= 0x40;
		}
	} else {
		NeoInput[2] &= ~0xF0;
	}

#if 1
	{
		nCyclesTotal[0] = (int)((long long)12000000 * nBurnCPUSpeedAdjust / (256 * 60));
		nCyclesScanline = nCyclesTotal[0] / 264;

		nNeoCPUSpeedfactor = (int)((long long)12000000 * nBurnCPUSpeedAdjust * 16384 / (4000000 * 256));
	}
#else
	nCyclesTotal[0] = 12000000 / 60;								// 12 MHz 68000
#endif

	nCyclesTotal[1] =  4000000 / 60;								//  4 MHz Z80

	bRenderImage = false;
	bForceUpdateOnStatusRead = false;

	if (pBurnDraw) {
		NeoUpdatePalette();											// Update the palette
		NeoClearScreen();
	}
	nSliceEnd = 0x10;

	SekOpen(0);
	ZetOpen(0);

	for (int i = 0; i < nInterleave; i++) {

		// Run 68000
		nCurrentCPU = 0;

		bIRQ2Enabled |= bRenderLineByLine || (nIRQ2Control & 0x10);
//		bIRQ2Enabled &= bAllowRasters;

		if (!bIRQ2Enabled) {
			nCyclesSegment = (i + 1) * nCyclesScanline * 0x18;
			nCyclesDone[0] += SekRun(nCyclesSegment - nCyclesDone[0]);
		} else {
			for (int j = 0; j < 0x18; j++) {
				int line = i * 0x18 + j;

				if ((nIRQ2Control & 0x10) && line == nIRQLine) {
					nIRQAcknowledge &= ~2;
					SekSetIRQLine(2, 1);
//					printf("  - IRQ triggered (line %3i).\n", line);

					if (nIRQ2Control & 0x80) {
						nIRQLine += NeoConvertIRQPositionRelative();
//						printf("  - IRQ Line -> %3i (at line %3i, autoload).\n", nIRQLine, line);
					}

					bForcePartialRender = bRenderImage;
				}

				if (bForcePartialRender) {
					nSliceStart = nSliceEnd;
					nSliceEnd = line - 7;
					nSliceSize = nSliceEnd - nSliceStart;

//					printf(" -- Drawing slice: %3i - %3i.\n", nSliceStart, nSliceEnd);
					NeoRenderSprites();							// Render sprites
				}
				bForcePartialRender = false;

				nCyclesSegment = (line + 1) * nCyclesScanline;
				nCyclesDone[0] += SekRun(nCyclesSegment - nCyclesDone[0]);
			}
		}

		{
			// Update timers
			if (pBurnSoundOut) {
				int nSegmentLength = nBurnSoundLen / (nInterleave + 1);
				dTime += (double)nSegmentLength / nBurnSoundRate;
			} else {
				dTime += 100.0 / (double)nBurnFPS / (double)(nInterleave + 1);
			}

			// Run Z80
			nCurrentCPU = 1;
			nCyclesSegment = (i + 1) * nCyclesTotal[1] /  (nInterleave + 1);
			if (BurnFMTimerUpdate(MAKE_TIMER_TICKS(nCyclesSegment - nCyclesDone[1], 4000000))) {
				ZetRaiseIrq(0xFF);
			}
			if ((nCyclesDone[1] - 0x0400) >= nCyclesSegment) {
				nCyclesDone[1] += ZetRun(0x0400);
			} else {
				nCyclesDone[1] += ZetRun(nCyclesSegment - nCyclesDone[1]);
			}

			// Render sound segment
			if (pBurnSoundOut) {
				int nSegmentLength = nBurnSoundLen / (nInterleave + 1);

				BurnYM2610Render(pBurnSoundOut + (nSoundBufferPos << 1), nSegmentLength);
				nSoundBufferPos += nSegmentLength;
			}
		}

		if (nTriggerIRQ == 1) {
			ZetRaiseIrq(0xFF);
			nCyclesDone[1] += ZetRun(0x0400);
		}
		nTriggerIRQ = 0;

		if (i == 0) {
			bRenderImage = pBurnDraw != NULL && bNeoEnableGraphics;
			bForceUpdateOnStatusRead = bRenderImage && bRenderLineByLine;
		}
	}

	bIRQ2Enabled |= bRenderLineByLine || (nIRQ2Control & 0x10);
//	bIRQ2Enabled &= bAllowRasters;

	if (!bIRQ2Enabled) {
		nCyclesSegment += nCyclesScanline * 0x08;
		nCyclesDone[0] += SekRun(nCyclesSegment - nCyclesDone[0]);
	} else {
		for (int j = 0; j < 8; j++) {
			int line = 240 + j;

			if ((nIRQ2Control & 0x10) && line == nIRQLine) {
				nIRQAcknowledge &= ~2;
				SekSetIRQLine(2, 1);
//				printf("  - IRQ triggered (line %3i).\n", line);

				if (nIRQ2Control & 0x80) {
					nIRQLine += NeoConvertIRQPositionRelative();
//					printf("  - IRQ Line -> %3i (at line %3i, autoload).\n", nIRQLine, line);
				}

				bForcePartialRender = bRenderImage;
			}

			if (bForcePartialRender) {
				bForcePartialRender = false;

				nSliceStart = nSliceEnd;
				nSliceEnd = line - 7;
				nSliceSize = nSliceEnd - nSliceStart;

//				printf(" -- Drawing slice: %3i - %3i.\n", nSliceStart, nSliceEnd);
				NeoRenderSprites();									// Render sprites
			}

			nCyclesSegment = (line + 1) * nCyclesScanline;
			nCyclesDone[0] += SekRun(nCyclesSegment - nCyclesDone[0]);
		}
	}

	if (bRenderImage) {
		if (nSliceEnd < 240) {
			nSliceStart = nSliceEnd;
			nSliceEnd = 240;
			nSliceSize = nSliceEnd - nSliceStart;

//			printf(" -- Drawing slice: %3i - %3i.\n", nSliceStart, nSliceEnd);
			NeoRenderSprites();										// Render sprites
		}
		NeoRenderText();											// Render text layer
	}

	nCurrentCPU = 0;

	nIRQAcknowledge &= ~4;
	SekSetIRQLine(1, 1);

//	printf("  - VBLank.\n");

	bIRQ2Enabled = bPrevIRQ2Enabled || bRenderLineByLine || (nIRQ2Control & 0x10);
//	bIRQ2Enabled &= bAllowRasters;

	bPrevIRQ2Enabled = bRenderLineByLine || (nIRQ2Control & 0x10);

	// (re-)set IRQ scanline after line 248 is done
	if (nIRQ2Control & 0x40) {
		nIRQLine = NeoConvertIRQPosition() - nIRQOffset;
//		printf("  - IRQ Line -> %3i (at line %3i, VBlank).\n", nIRQLine, 248);
	}

	if (!bIRQ2Enabled) {
		nCyclesSegment = nCyclesTotal[0];
		nCyclesDone[0] += SekRun(nCyclesSegment - nCyclesDone[0]);
	} else {
		for (int j = 0; j < 16; j++) {
			int line = 248 + j;
			if ((nIRQ2Control & 0x10) && line == nIRQLine) {
				nIRQAcknowledge &= ~2;
				SekSetIRQLine(2, 1);
//				printf("  - IRQ triggered   (at line %3i).\n", line);

				if (nIRQ2Control & 0x80) {
					nIRQLine += NeoConvertIRQPositionRelative();
//					printf("  - IRQ Line -> %3i (at line %3i, autoload).\n", nIRQLine, line);
				}
			}

			nCyclesSegment = (line + 1) * nCyclesScanline;
			nCyclesDone[0] += SekRun(nCyclesSegment - nCyclesDone[0]);
		}
	}

	if (nIRQLine > 263) {
//		printf(" -- IRQ Line past screen end (nIRQ2Control: %02X, nIRQLine: %3i).\n", nIRQ2Control, nIRQLine);
//		printf(" -- IRQ Line past screen end (nIRQLineHi: %04X, nIRQLineLo: %04X).\n", nIRQLineHi, nIRQLineLo);
		if ((nIRQ2Control & 0x20)) {
			nIRQLine -= 264;
		} else {
			if (!(nIRQ2Control & 0x10)) {
				nIRQLine = 0;
			}
		}
	}
//	printf(" -- nIRQ2Control: %02X, nIRQLine: %3i.\n", nIRQ2Control, nIRQLine);

	{
		// Render sound segment
		if (pBurnSoundOut) {
			int nSegmentLength = nBurnSoundLen - nSoundBufferPos;
			if (nSegmentLength) {
				dTime = dT + (double)nSegmentLength / nBurnSoundRate;
			}
		} else {
			dTime = dT + 100.0 / (double)nBurnFPS;
		}

		nCurrentCPU = 1;
		nCyclesSegment = nCyclesTotal[1];
		if (BurnFMTimerUpdate(MAKE_TIMER_TICKS(nCyclesSegment - nCyclesDone[1], 4000000))) {
			ZetRaiseIrq(0xFF);
			nCyclesDone[1] += ZetRun(0x0400);
		}
		nCyclesDone[1] += ZetRun(nCyclesTotal[1] - nCyclesDone[1]);

		if (pBurnSoundOut) {
			int nSegmentLength = nBurnSoundLen - nSoundBufferPos;
			if (nSegmentLength) {
				BurnYM2610Render(pBurnSoundOut + (nSoundBufferPos << 1), nSegmentLength);
				nSoundBufferPos += nSegmentLength;
			}
		}
	}

	if (nTriggerIRQ == 1) {
		ZetRaiseIrq(0xFF);
		nCyclesDone[1] += ZetRun(0x0400);
	}
	nTriggerIRQ = 0;

	uPD4990AUpdate(nCyclesDone[0] - nuPD4990ATicks);

	// Compensate for extra cycles executed
	nCyclesDone[0] -= nCyclesTotal[0];
	nCyclesDone[1] -= nCyclesTotal[1];

	ZetClose();
	SekClose();

	if ((nIRQ2Control & 8) == 0) {
		if (++nSpriteFrameTimer > nSpriteFrameSpeed) {
			nSpriteFrameTimer = 0;
			nNeoSpriteFrame++;
		}
	}

	dTime = dT + 1.0 / 60.0;

	return 0;
}
